/* ******************************************************************* *\
*                               bitarray.d                              *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*   Data types to implement bit pointers and arrays with a bit offset   *
>***********************************************************************<
* Copyright (C) 2005-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
The smjg.libs.util.bitarray module provides data types to represent bit pointers and bit arrays
that don't necessarily begin on byte boundaries.

$(P In terms of the internal representation, bits are indexed within bytes in a way that corresponds
to the platform byte order.  So on little-endian platforms, bit 0 is the least significant bit,
and bit 7 the most significant.  On big-endian platforms, the convention is reversed.  This is
significant if a bit array is constructed from or converted to an integer array, and has the
effect that the indexing order of bits matches (in one direction or the other) the significance
order of bits in the integers involved.)

Todo:
	Test on big-endian platforms.

Bugs:
	Not yet supported on 64-bit platforms.
*/
module smjg.libs.util.bitarray;
private import std.traits;

version (D_LP64) {
	static assert (false, "smjg.libs.util.bitarray not supported on 64-bit platforms");
}

private const int
	BIT_OFFSET_SHIFT	= (uint.sizeof << 3) - 3,
	MIN_ALLOC_SIZE		= 4;

/// The BitPointer structure implements a bit pointer that can point to any bit within a byte.
struct BitPointer {
	private {
		union {
			struct {
				version (LittleEndian) {
					uint _bitOffset;
					ubyte* _bytePointer;
				} else {
					ubyte* _bytePointer;
					uint _bitOffset;
				}
			}
			ulong _wholeThing;
		}
	}

	static assert (ulong.sizeof == uint.sizeof + (ubyte*).sizeof);

	const invariant() {
		assert ((_bitOffset & ~((-1) << BIT_OFFSET_SHIFT)) == 0);
	}

	/// Constructs a new BitPointer to point to the beginning of a BitArray.
	this(inout(BitArray) data) inout {
		_wholeThing = data.ptr._wholeThing;
	}

	///	Constructs a new BitPointer to point to the first bit of a byte.
	this(inout(void)* data) inout {
		_bytePointer = cast(inout(ubyte)*) data;
	}

	version (LittleEndian) {
		/// Dereferences the bit pointer.
		@property bool value() const {
			return cast(bool)
			  ((*_bytePointer >>> (_bitOffset >>> BIT_OFFSET_SHIFT)) & 1);
		}

		/// ditto
		@property bool value(bool b) {
			*_bytePointer &= ~(1 << (_bitOffset >>> BIT_OFFSET_SHIFT));
			*_bytePointer |= (b << (_bitOffset >>> BIT_OFFSET_SHIFT));
			return b;
		}
	} else {
		/// Dereferences the bit pointer.
		@property bool value() const {
			return cast(bool) ((*_bytePointer >>>
			    (7 - (_bitOffset >>> BIT_OFFSET_SHIFT))) & 1);
		}

		/// ditto
		@property bool value(bool b) {
			*_bytePointer &= ~(1 << (7 - (_bitOffset >>> BIT_OFFSET_SHIFT)));
			*_bytePointer |= (b << (7 - (_bitOffset >>> BIT_OFFSET_SHIFT)));
			return b;
		}
	}

	///	Accesses the bit at position i relative to this bit pointer.
	bool opIndex(int i) const {
		return opAdd(i).value;
	}

	/// ditto
	bool opIndexAssign(bool b, int i) {
		return opAdd(i).value = b;
	}

	/**	Accesses the specified range of bits relative to this bit pointer.
	 *
	 *	Todo: optimise assignment.
	 */
	inout(BitArray) opSlice(int begin, int end) inout {
		return inout(BitArray)(opAdd(begin), end - begin);
	}

	/// ditto
	BitArray opSliceAssign(bool value, int begin, int end) {
		BitArray slice = opSlice(begin, end);

		foreach (ref bool b; slice) {
			b = value;
		}
		return slice;
	}

	/// ditto
	BitArray opSliceAssign(in BitArray value, int begin, int end)
	in {
		// check lengths match
		assert (end - begin == value._length);

		// check slices don't overlap
		assert (opAdd(end) <= value.ptr
		  || value.ptr + value._length <= opAdd(begin));
	}
	out {
		assert (opSlice(begin, end) == value);
	}
	body {
		BitArray slice = opSlice(begin, end);
		// *** optimise
		foreach (int index, ref bool b; slice) {
			b = value[index];
		}
		return slice;
	}

	///	Returns: this pointer incremented by i bits.
	inout(BitPointer) opAdd(int i) inout {
		return inout(BitPointer)(_wholeThing + (cast(ulong) i << BIT_OFFSET_SHIFT));
	}

	///	Returns: this pointer decremented by i bits.
	inout(BitPointer) opSub(int i) inout {
		return inout(BitPointer)(_wholeThing - (cast(ulong) i << BIT_OFFSET_SHIFT));
	}

	///	Returns: the displacement in bits of this BitPointer from p.
	int opSub(in BitPointer p) const {
		return cast(int) ((_wholeThing - p._wholeThing) >> BIT_OFFSET_SHIFT);
	}

	/// Compares this BitPointer with another in the address to which they point.
	int opCmp(in BitPointer p) const { return opSub(p); }

	/// Returns: the address of the byte within which this BitPointer points.
	inout(ubyte)* opCast(T : inout(ubyte)*)() inout {
		return _bytePointer;
	}

	/** Returns: the address of the byte within which this BitPointer points.
	 *	Since: 0.03
	 */
	@property inout(ubyte)* bytePointer() inout {
		return _bytePointer;
	}

	/**	Returns: the bit offset of the bit pointed to within its byte, which is between 0 and 7.
	 *	Since: 0.03
	 */
	@property ubyte bitOffset() const {
		return _bitOffset >>> BIT_OFFSET_SHIFT;
	}

	private this(ulong wholeThing) inout {
		_wholeThing = wholeThing;
	}
}


///	The BitArray structure implements a bit array that can start at any point within a byte.
struct BitArray {
	private {
		BitPointer _start;
		int _length;
		int _allocBytes;
	}

	invariant() {
		// will be 0 if a slice, otherwise must be big enough
		assert (_allocBytes == 0 || _length <= (_allocBytes << 3));
	}

	/// Constructs a new BitArray with the length len.
	this(int len) {
		static const int BLOCK_SIZE_EXP	= 11;	// 2k blocks
		_length = len;

		// number of whole bytes used (excluding final byte)
		int wholeBytes = ((len - 1) >> 3);

		// actual number of bytes to allocate
		if (wholeBytes >= (1 << (BLOCK_SIZE_EXP - 1))) {
			// more than half a block - allocate a whole number of blocks
			_allocBytes
			  = (((wholeBytes+1) >> BLOCK_SIZE_EXP) + 1) << BLOCK_SIZE_EXP;
		} else {
			// allocate a power of 2 bytes
			for (_allocBytes = MIN_ALLOC_SIZE; _allocBytes <= wholeBytes + 1;
			  _allocBytes <<= 1) {}
		}

		debug (BitArray) writefln("alloc %d/%d", _allocBytes, len);
		_start._bytePointer = (new ubyte[_allocBytes]).ptr;
	}

	/// Constructs a new BitArray from an array of booleans.
	this(in bool[] bits) {
		this(bits.length);

		foreach (uint index, bool b; bits) {
			this[index] = b;
		}
	}

	/// Constructs a new BitArray to wrap a D array.
	this(inout(void[]) data) inout {
		_length = data.length * 8;
		_start = inout(BitPointer)(data.ptr);
	}

	/**	Compares two BitArrays for equality of the data they contain.
	 *	Todo: optimise for arrays byte-aligned with each other.
	 */
	int opEquals(in BitArray ba) const {
		if (this is ba) return true;
		if (ba._length != _length) return false;

		for (int index = 0; index < _length; index++) {
			if (opIndex(index) != ba[index]) return false;
		}

		return true;
	}

	///	Accesses the bit at position i.
	bool opIndex(int i) const
	in {
		assert (i >= 0);
		assert (i < _length);
	}
	body {
		return _start[i];
	}

	/// ditto
	bool opIndexAssign(bool b, int i)
	in {
		assert (i >= 0);
		assert (i < _length);
	}
	body {
		return _start[i] = b;
	}

	///	Accesses the specified slice of this array.
	inout(BitArray) opSlice(int begin, int end) inout
	in {
		checkBounds(begin, end);
	}
	body {
		return _start[begin..end];
	}

	/// ditto
	BitArray opSliceAssign(bool value, int begin, int end)
	in {
		checkBounds(begin, end);
	}
	body {
		return _start.opSliceAssign(value, begin, end);
	}

	/// ditto
	BitArray opSliceAssign(in BitArray value, int begin, int end)
	in {
		checkBounds(begin, end);
	}
	body {
		return _start.opSliceAssign(value, begin, end);
	}

	///	Assigns value to all elements of this BitArray.
	BitArray opSliceAssign(bool value) {
		return _start.opSliceAssign(value, 0, _length);
	}

	///	Assigns a BitArray in-place to this one.
	BitArray opSliceAssign(in BitArray value) {
		return _start.opSliceAssign(value, 0, _length);
	}

	///	Concatenates two BitArray objects.
	inout(BitArray) opCat(in BitArray value) inout {
		BitArray ba = BitArray(_length + value._length);
		ba.opSliceAssign(this, 0, _length);
		ba.opSliceAssign(value, _length, _length + value._length);
		return cast(inout) ba;
	}

	///	Concatenates this BitArray with a bit.
	inout(BitArray) opCat(bool value) inout {
		BitArray ba = BitArray(_length + 1);
		ba.opSliceAssign(this, 0, _length);
		ba[_length] = value;
		return cast(inout) ba;
	}

	///	Appends value to this BitArray.
	BitArray opCatAssign(in BitArray value) {
		int catAt = length;
		length = catAt + value.length;
		opSliceAssign(value, catAt, catAt + value.length);
		return this;
	}

	/// ditto
	BitArray opCatAssign(bool value) {
		length = length + 1;
		opIndexAssign(value, _length - 1);
		return this;
	}

	/// A BitPointer pointing to the first element of the array.
	@property inout(BitPointer) ptr() inout {
		return _start;
	}

	///	Iterates over the elements of this BitArray.
	int opApply(int delegate(ref bool) dg) {
		int result = 0;

		for (int i = 0; i < _length; i++) {
			bool b = _start[i];
			result = dg(b);
			_start[i] = b;
			if (result) break;
		}
		return result;
	}

	/// ditto
	int opApply(int delegate(bool) dg) const {
		int result = 0;

		for (int i = 0; i < _length; i++) {
			result = dg(_start[i]);
			if (result) break;
		}
		return result;
	}

	/// ditto
	int opApply(int delegate(int, ref bool) dg) {
		int result = 0;

		for (int i = 0; i < _length; i++) {
			bool b = _start[i];
			result = dg(i, b);
			_start[i] = b;
			if (result) break;
		}
		return result;
	}

	/// ditto
	int opApply(int delegate(int, bool) dg) const {
		int result = 0;

		for (int i = 0; i < _length; i++) {
			result = dg(i, _start[i]);
			if (result) break;
		}
		return result;
	}

	///	Inverts the bits of the array.
	inout(BitArray) opUnary(string Op : "~")() inout {
		if (_length == 0) return BitArray.init;

		BitArray result = dup;
		BitPointer end = result._start + _length;
		ubyte* lastByte = end._bytePointer + (end._bitOffset != 0);
		ubyte* b;

		for (b = result._start._bytePointer; b < lastByte; b++) *b = ~*b;
		return cast(inout) result;
	}

	/**	Modifies this array by performing a bitwise operation with ba.
	 *	Todo: optimise
	 */
	BitArray opOpAssign(string Op)(in BitArray ba) if (Op == "&" || Op == "|" || Op == "^")
	in {
		assert (_length == ba.length);
	}
	body {
		foreach (i, ref b; this) {
			mixin("b " ~ Op ~ "= ba[i];");
		}
		return this;
	}

	/**	Modifies this array by performing a bitwise AND-NOT operation with ba.
	 *	Todo: optimise
	 */
	BitArray opOpAssign(string Op)(in BitArray ba) if (Op == "-")
	in {
		assert (_length == ba.length);
	}
	body {
		foreach (i, ref b; this) {
			b &= !ba[i];
		}
		return this;
	}

	///	Performs a bitwise operation on this array and ba.
	inout(BitArray) opBinary(string Op)(in BitArray ba) inout {
		return cast(inout) dup.opOpAssign!(Op)(ba);
	}

	///	Returns: a byte-aligned copy of this array.
	@property BitArray dup() const
	out (result) {
		assert (this == result);
	}
	body {
		BitArray p = BitArray(_length);
		p.opSliceAssign(this);
		return p;
	}

	///	Returns: an immutable, byte-aligned copy of this array.
	@property immutable(BitArray) idup() const {
		return cast(immutable) dup();
	}

	///	Returns: this BitArray, if it is already byte aligned, otherwise a byte-aligned copy.
	@property inout(BitArray) byteAlign() inout {
		if (_start._bitOffset == 0) {
			return this;
		} else {
			return cast(inout) dup;
		}
	}

	///	Returns: the _length in bits of this BitArray.
	@property int length() const { return _length; }

	/// ditto
	@property int length(int len)
	in {
		assert (len >= 0);
	}
	body {
		if (len > _length && len > (_allocBytes << 3)) {
			// too big - need to reallocate
			BitArray reallocate = BitArray(len);

			// copy data from this
			reallocate.opSliceAssign(this, 0, _length);
			this = reallocate;

		} else {
			_length = len;
		}
		return len;
	}

	///	Returns: a string representation of this BitArray.
	@property string toString() {
		return toStringBySpec('0', '1');
	}

	/**	Creates a custom string representation of this array.
	 *	Params:
	 *		falseAs = character used to represent a false (0) byte.
	 *		trueAs  = character used to represent a true (1) byte.
	 */
	string toStringBySpec(char falseAs, char trueAs) {
		// redundant, but included for efficiency
		char[] result;
		result.length = length;
		foreach (i, b; this) result[i] = b ? trueAs : falseAs;
		return cast(string) result;
	}

	/// ditto
	string toStringBySpec(dchar falseAs, dchar trueAs) {
		char[] result;
		foreach (i, b; this) result ~= b ? trueAs : falseAs;
		return cast(string) result;
	}

	/**	Converts this BitArray _to an integer array.
	 *
	 *	$(P to uses the native bit order; toRev uses the reverse of this.)
	 *
	 *	$(P All of these functions return an in-place view of the bit array if possible (i.e. native
	 *	bit order, and begins and ends on byte boundaries), otherwise a copy.)
	 */
	@property inout(U)[] to(T : U[], U)() inout if (isIntegral!U) {
		return intArrayConv!(U).arrayFrom(this);
	}

	/// ditto
	@property inout(U)[] toRev(T : U[], U)() inout if (isIntegral!U) {
		return intArrayConv!(U).revArrayFrom(this);
	}

	version (LittleEndian) {
		/// ditto
		alias to toLittleEndian;
		/// ditto
		alias toRev toBigEndian;
	} else {
		/// ditto
		alias toRev toLittleEndian;
		/// ditto
		alias to toBigEndian;
	}

	deprecated {
		alias to!(ubyte[]) toUbyteArray;
		alias to!(ushort[]) toUshortArray;
		alias to!(uint[]) toUintArray;
		alias to!(ulong[]) toUlongArray;
		alias toLittleEndian!(ubyte[]) toLEUbyteArray;
		alias toLittleEndian!(ushort[]) toLEUshortArray;
		alias toLittleEndian!(uint[]) toLEUintArray;
		alias toLittleEndian!(ulong[]) toLEUlongArray;
		alias toBigEndian!(ubyte[]) toBEUbyteArray;
		alias toBigEndian!(ushort[]) toBEUshortArray;
		alias toBigEndian!(uint[]) toBEUintArray;
		alias toBigEndian!(ulong[]) toBEUlongArray;
	}

	private {
		this(inout(BitPointer) start, int length) inout {
			_start = start;
			_length = length;
			_allocBytes = 0;
		}

		void checkBounds(int begin, int end) const {
			assert (begin >= 0);
			assert (begin <= end);
			assert (end <= _length);
		}
	}
}


private template intArrayConv(T) {
	// convert to integer array with native bit order
	inout(T)[] arrayFrom(inout(BitArray) ba) {
		if (ba.ptr._bitOffset == 0 && ba.length % (T.sizeof * 8) == 0) {
			// byte aligned and ready - return it in place
			return cast(inout(T)[]) ba.ptr._bytePointer[0 .. ba.length / 8];
		} else {
			BitArray newBA = makeMultiple(ba);
			return cast(inout(T)[]) newBA.ptr._bytePointer[0 .. newBA.length / 8];
		}
	}

	// convert to integer array with foreign bit order
	inout(T)[] revArrayFrom(inout(BitArray) ba) {
		BitArray newBA = makeMultiple(ba);
		T[] result = cast(T[]) newBA.ptr._bytePointer[0 .. newBA.length / 8];

		foreach (ref T element; result) {
			element = reverseBits(element);
		}
		return cast(inout) result;
	}

	// make a copy of the array rounded up to a multiple of this type's size
	static if (T.sizeof <= MIN_ALLOC_SIZE) {
		BitArray makeMultiple(in BitArray ba) {
			BitArray result = ba.dup;
			if (ba.length % (T.sizeof * 8) != 0) {
				result.length = (ba.length / (T.sizeof * 8) + 1)
				  * (T.sizeof * 8);
			}
			return result;
		}
	} else {
		BitArray makeMultiple(in BitArray ba) {
			int len = ba.length;
			if (len % (T.sizeof * 8) != 0) {
				len = (len / (T.sizeof * 8) + 1) * (T.sizeof * 8);
			}

			BitArray result = BitArray(len);
			result.opSliceAssign(ba, 0, ba.length);
			return result;
		}
	}

	static if (T.sizeof == 1) {
		T reverseBits(T value) {
			return (value >>> 7)
				| ((value & 0x40) >>> 5)
				| ((value & 0x20) >>> 3)
				| ((value & 0x10) >>> 1)
				| ((value & 0x08) << 1)
				| ((value & 0x04) << 3)
				| ((value & 0x02) << 5)
				| ((value & 0x01) << 7);
		}

	} else {
		T reverseBits(T value) {
			T result;
			ubyte* vb = cast(ubyte*) &value + T.sizeof;
			ubyte* rb = cast(ubyte*) &result;

			while (vb != cast(ubyte*) &value) {
				*(rb++) = intArrayConv!(ubyte).reverseBits(*--vb);
			}
			return result;
		}
	}
}


unittest {
	import std.stdio;

	static bool[] data = [1, 1, 0, 1, 0, 0, 0, 0];
	version (LittleEndian) {
		ubyte ub = 0b00001011;
	} else {
		ubyte ub = 0b11010000;
	}
	BitArray ba = BitArray(data);

	assert (BitArray((&ub)[0..1]) == ba);

	assert (ba[0] == 1);
	assert (ba[1] == 1);
	assert (ba[2] == 0);
	assert (ba[3] == 1);
	assert (ba[4] == 0);
	assert (ba[5] == 0);
	assert (ba[6] == 0);
	assert (ba[7] == 0);

	// indexing
	ba[0] = 0;
	assert (ba[0] == 0);
	ba[0] = 1;
	assert (ba[0] == 1);
	ba[1] = 0;
	assert (ba[1] == 0);
	ba[1] = 1;
	assert (ba[1] == 1);
	ba[4] = 1;
	assert (ba[4] == 1);
	ba[4] = 0;
	assert (ba[4] == 0);
	ba[4] = ba[1];
	assert (ba[1] == ba[4]);
	ba[4] = ba[2];
	assert (ba[2] == ba[4]);

	ba = BitArray(data[0..5]);

	assert (ba.length == 5);
	assert (ba[0] == 1);
	assert (ba[1] == 1);
	assert (ba[2] == 0);
	assert (ba[3] == 1);
	assert (ba[4] == 0);

	// slicing
	BitArray slice = ba[2..4];

	slice[0] = 1;
	assert (ba[2] == 1);

	ba[2] = 0;
	assert (slice[0] == 0);

	puts("BitArray basic ops passed!");

	BitPointer p = ba.ptr;

	// pointing
	p.value = 0;
	assert (ba[0] == 0);
	p.value = 1;
	assert (ba[0] == 1);

	// indexing
	p[0] = 0;
	assert (ba[0] == 0);
	p[0] = 1;
	assert (ba[0] == 1);
	p[1] = 0;
	assert (ba[1] == 0);
	p[1] = 1;
	assert (ba[1] == 1);
	p[4] = 1;
	assert (ba[4] == 1);
	p[4] = 0;
	assert (ba[4] == 0);
	p[4] = p[1];
	assert (ba[1] == ba[4]);
	p[4] = p[2];
	assert (ba[2] == ba[4]);

	assert (p[0] == 1);
	assert (p[1] == 1);
	assert (p[2] == 0);
	assert (p[3] == 1);
	assert (p[4] == 0);

	// pointer arithmetic
	BitPointer q = p + 3;
	BitPointer o = q - 2;
	assert (o == p + 1);

	o[0] = 0;
	assert (p[1] == 0);
	p[1] = 1;
	assert (o[0] == 1);

	puts("BitPointer unit test passed!");

	string textRep;


	// foreach - in unindexed
	foreach (bool b; ba) {
		//puts(b ? "true " : "false ");
		textRep ~= (b ? 't' : 'f');
	}
	assert (textRep == "ttftf");

	// foreach - ref unindexed
	foreach (ref bool b; ba) {
		b = !b;
	}

	textRep = "";

	// foreach - in indexed
	foreach (int i, bool b; ba) {
		textRep ~= cast(char) ('0' + i);
		textRep ~= (b ? 't' : 'f');
		//printf("Element %d: %d\n", i, cast(int) b);
	}
	assert (textRep == "0f1f2t3f4t");

	textRep = "";

	// foreach - ref indexed
	foreach (int i, ref bool b; ba) {
		textRep ~= cast(char) ('0' + i);
		textRep ~= (b ? 't' : 'f');
		b = !b;
		textRep ~= (b ? 't' : 'f');
	}

	assert (textRep == "0ft1ft2tf3ft4tf");
	assert (ba.toString == "11010");

	puts("BitArray iteration passed!");

	ba[1..4] = 0;
	foreach (bool b; ba[1..4]) assert (b == 0);

	ba[1..4] = 1;
	foreach (bool b; ba[1..4]) assert (b == 1);
	assert (ba[0] == 1);
	assert (ba[4] == 0);

	puts("bool to BitArray slice assignment passed!");
}


unittest {
	import std.stdio;
	import core.exception;

	static bool[] data = [0,0,0,1,0,0,1,1, 0,1,0,0,1,1,1,1];
	BitArray ba = BitArray(data);
	BitArray slice1 = ba[1..7];
	BitArray slice2 = ba[8..14];
	BitArray slice3 = ba[9..14];

	ba[1..7] = slice2;

	assert (slice1 == slice2);
	assert (slice1 != ba[9..14]);
	assert (slice1 != ba[9..15]);
	assert (slice1 !is slice2);
	assert (slice1 is ba[1..7]);

	try {
		ba[4..8] = ba[5..9];
		throw new Error("Slice overlap validation failed!");
	} catch (AssertError e) {}

	puts("BitArray slice copying and comparison passed!");

	BitArray newData = ba;
	newData.length = 30;
	assert (newData[0..16] == ba);
	newData.length = 2049;
	assert (newData[0..16] == ba);


	puts("BitArray length setting passed!");

	newData = ba[2..5] ~ ba[10..16];
	assert (newData[0..3] == ba[2..5]);
	assert (newData[3..9] == ba[10..16]);

	newData ~= ba[8..13];
	newData ~= 0;
	newData ~= 1;
	assert (newData[9..14] == ba[8..13]);
	assert (newData[14] == 0);
	assert (newData[15] == 1);

	puts("BitArray concatenation passed!");

	ba = BitArray(data);

	version (LittleEndian) {
		static ubyte[] bytes = [ 0b11001000, 0b11110010 ];
	} else {
		static ubyte[] bytes = [ 0b00010011, 0b01001111 ];
	}
	assert (ba == BitArray(bytes));

	puts("BitArray construction from ubyte array passed!");

	assert (ba.toString == "0001001101001111");
	ba ~= 1;
	ba ~= 0;
	assert (ba.toString == "000100110100111110");

	puts("BitArray conversion to string passed!");
}

unittest {
	import std.stdio;

	static bool[] data1 = [ 0,0,0,1,0,0,1,1, 0,1,0,0,1,1,1,1 ];
	static bool[] data2 = [ 1,1,0,1,0,0,1,0, 1,0,1,0,1,1,0,1 ];
	BitArray ba1 = BitArray(data1);
	BitArray ba2 = BitArray(data2);

	BitArray result = ba1 & ba2;
	assert(result.toString == "0001001000001101");
	result = ba1 | ba2;
	assert(result.toString == "1101001111101111");
	result = ba1 ^ ba2;
	assert(result.toString == "1100000111100010");
	result &= ba1;
	assert(result.toString == "0000000101000010");
	result |= ba2;
	assert(result.toString == "1101001111101111");
	result = ba1;
	result ^= ba2;
	assert(result.toString == "1100000111100010");

	puts("BitArray bitwise operations passed!");
}

unittest {
	import std.stdio;

	static bool[] data = [
		0,0,0,1,0,0,1,1, 0,1,0,0,1,1,1,1, 1,1,0,1,0,0,1,0, 1,0,1,0,1,1,0,1];
	BitArray ba = BitArray(data);

	ubyte[] ub = ba.toLittleEndian!(ubyte[]);
	ushort[] us = ba.toLittleEndian!(ushort[]);
	uint[] ui = ba.toLittleEndian!(uint[]);
	ulong[] ul = ba.toLittleEndian!(ulong[]);

	assert (ub.length == 4);
	assert (ub[0] == 0b11001000);
	assert (ub[1] == 0b11110010);
	assert (ub[2] == 0b01001011);
	assert (ub[3] == 0b10110101);

	assert (us.length == 2);
	assert (us[0] == 0b1111001011001000);
	assert (us[1] == 0b1011010101001011);

	assert (ui.length == 1);
	assert (ui[0] == 0b10110101010010111111001011001000);

	assert (ul.length == 1);
	assert (ul[0] == 0b10110101010010111111001011001000);

	ub = ba.toBigEndian!(ubyte[]);
	us = ba.toBigEndian!(ushort[]);
	ui = ba.toBigEndian!(uint[]);
	ul = ba.toBigEndian!(ulong[]);

	assert (ub.length == 4);
	assert (ub[0] == 0b00010011);
	assert (ub[1] == 0b01001111);
	assert (ub[2] == 0b11010010);
	assert (ub[3] == 0b10101101);

	assert (us.length == 2);
	assert (us[0] == 0b0001001101001111);
	assert (us[1] == 0b1101001010101101);

	assert (ui.length == 1);
	assert (ui[0] == 0b00010011010011111101001010101101);

	assert (ul.length == 1);
	assert (ul[0] == (0b00010011010011111101001010101101UL << 32));

	ba = ba[5..28];

	ub = ba.toLittleEndian!(ubyte[]);
	us = ba.toLittleEndian!(ushort[]);
	ui = ba.toLittleEndian!(uint[]);
	ul = ba.toLittleEndian!(ulong[]);

	assert (ub.length == 3);
	assert (ub[0] == 0b10010110);
	assert (ub[1] == 0b01011111);
	assert (ub[2] == 0b00101010);

	assert (us.length == 2);
	assert (us[0] == 0b0101111110010110);
	assert (us[1] == 0b00101010);

	assert (ui.length == 1);
	assert (ui[0] == 0b001010100101111110010110);

	assert (ul.length == 1);
	assert (ul[0] == 0b001010100101111110010110);

	ub = ba.toBigEndian!(ubyte[]);
	us = ba.toBigEndian!(ushort[]);
	ui = ba.toBigEndian!(uint[]);
	ul = ba.toBigEndian!(ulong[]);
// 01101001 11111010 0101010

	assert (ub.length == 3);
	assert (ub[0] == 0b01101001);
	assert (ub[1] == 0b11111010);
	assert (ub[2] == 0b01010100);

	assert (us.length == 2);
	assert (us[0] == 0b0110100111111010);
	assert (us[1] == 0b0101010000000000);

	assert (ui.length == 1);
	assert (ui[0] == 0b01101001111110100101010000000000);

	assert (ul.length == 1);
	assert (ul[0] == 0b01101001111110100101010000000000UL << 32);

	puts("BitArray conversion to integer arrays passed!");
}
