/* ******************************************************************* *\
*                             commandline.d                             *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*   Conversion of command lines between single string and string array  *
*                                 forms                                 *
>***********************************************************************<
* Copyright (C) 2005-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
The smjg.libs.util.commandline module provides functions to construct command lines from argument
arrays and vice versa.

$(P These functions involve conversion of arguments between two forms:)
$(UL
$(LI "application form" - the form received by the main function in C and D, and the form that the
                          application should use)
$(LI "command line form" - the form that is actually typed on the command line.)
)
Since: 0.02
*/
module smjg.libs.util.commandline;

private import
	std.array,
	std.string,
	std.utf;
debug (commandline) private import std.stdio;

version (Windows) pragma(lib, "shell32.lib");


/**	Builds a command line string from an array of arguments.
 *
 *	Params:
 *		args = array of arguments in application form.
 *	Returns:
 *		a representation of the arguments that can be entered on the OS command line.
 */
string buildCommandLine(const(string)[] args) {
	string line;

	foreach (index, arg; args) {
		line ~= formatArgument(arg);
		if (index != args.length - 1) line ~= ' ';
	}

	return line;
}


/**	Converts a null-terminated UTF-16 string into a UTF-8 D string.
 *
 *	$(P This is used by the Windows implementation of the command line
 *	functions, hence its inclusion here.)
 *
 *	Params:
 *		s = a pointer to a null-terminated UTF-16 string.
 *	Returns:
 *		the string converted into UTF-8.
 */
string wcharzToUTF8(const(wchar)* s) {
	char[] result;

	while (*s != '\0') {
		encode(result, *s);
	}
	return cast(string) result;
}


version (Windows) {
	private import std.c.windows.windows;

	extern (Windows) LPWSTR GetCommandLineW();

	/**	Converts a single argument from application form to command line form.
	 *
	 *	Todo: finish implementing on platforms other than Windows.
	 */
	string formatArgument(string arg) {
		string from = `"`, to = `\"`;
		arg = replace(arg, from, to);

		// translate backslashes preceding a quotation mark
		int bsl = arg.length;
		while ((bsl = lastIndexOf(arg[0..bsl], `\\"`)) != -1) {
			int firstBsl;
			for (firstBsl = bsl++;
			  firstBsl >= 0 && arg[firstBsl] == '\\';
			  firstBsl--) {}
			firstBsl++;
			arg = arg[0..bsl] ~ arg[firstBsl..$];
			bsl = firstBsl;
		}

		if (indexOf(arg, ' ') != -1) {
			// enclose in quotes, but first double final backslashes
			if (arg[$-1] == '\\') {
				int firstBsl;
				for (firstBsl = arg.length - 1;
				     firstBsl >= 0 && arg[firstBsl] == '\\';
				     firstBsl--) {}
				firstBsl++;
				return `"` ~ arg ~ arg[firstBsl..$] ~ `"`;
			}

			return `"` ~ arg ~ `"`;
		}
		return arg;
	}

	/**	Parses a command _line into an argument list.
	 *
	 *	Params:
	 *		line = the command _line.
	 *	Returns:
	 *		an array of the arguments in application form.
	 *
	 *	Todo: implement on platforms other than Windows.
	 */
	string[] parseCommandLine(const(char)[] line) {
		string[] args;
		bool inQuotes;

		while (line.length != 0) {
			// skip leading whitespace
			while (line[0] == ' ') {
				line = line[1..$];
				if (line.length == 0) return args;
			}

			int nextSpace = indexOf(line, ' ');
			if (nextSpace == -1) nextSpace = line.length;

			int nextQuote = indexOf(line, '"');
			string argSoFar;

			// now read the argument
			// loop for as long as there's a '"' before the next ' '
			debug (commandline) writefln("So far: %s\nLeft: %s", argSoFar, line);

		readAnArg:
			while (nextQuote != -1 && nextQuote < nextSpace) {
				if (nextQuote == 0) {
					inQuotes = true;
					line = line[1..$];
				} else {
					const(char)[] portion = line[0..nextQuote];

					/*	count backslashes preceding '"'
					 *	if an odd number, a literal '"' is there, so a quoted
					 *	portion is not being opened
					 */
					inQuotes = !halveBackslashes(portion);
					argSoFar ~= portion;
					line = line[nextQuote+1..$];
				}

				nextQuote = indexOf(line, '"');
				// now read quoted portion
				while (inQuotes) {
					if (nextQuote != -1) {
						const(char)[] portion = line[0..nextQuote];

						/*	count backslashes preceding '"'
						 *	if an odd number, a literal '"' is there, so the
						 *	quoted portion is not being closed
						 */
						inQuotes = halveBackslashes(portion);
						argSoFar ~= portion;
						line = line[nextQuote+1..$];
						nextQuote = indexOf(line, '"');
					} else {
						// argument extends to end of string
						inQuotes = false;
						nextSpace = line.length;
						break readAnArg;
					}
				}
				nextSpace = indexOf(line, ' ');
				if (nextSpace == -1) nextSpace = line.length;
			}	// readAnArg

			version (D_Version2) {
				args ~= argSoFar ~ line[0..nextSpace].idup;
			} else {
				args ~= argSoFar ~ line[0..nextSpace].dup;
			}
			line = line[nextSpace..$];
		}

		return args;
	}

	private bool halveBackslashes(ref const(char)[] portion) {
		if (portion.length == 0 || portion[$-1] != '\\') return false;

		int index;
		for (index = portion.length - 2;
			  index >= 0 && portion[index] == '\\';
			  index--) {}
		// index is now one position before first '\\' of the string
		switch ((portion.length - index) % 2) {
			case 0:
				/*	odd number of backslashes
				 *	abc\\\\\ -> abc\\"
				 *	portion.length == 8
				 *	index == 2
				 *	desired length: 5 plus '"'
				 */
				portion = portion[0..($ + index)/2] ~ `"`;
				return true;

			case 1:
				/*	even number of backslashes
				 *	abc\\\\\\ -> abc\\\
				 *	portion.length == 9
				 *	index == 2
				 *	desired length: 6
				 */
				portion = portion[0..($ + index + 1)/2];
				return false;

			default:
				assert (false);
		}
	}

	/**	Parses the current application's command line into an argument list.
	 *
	 *	Returns:
	 *		an array of the arguments in application form.
	 *		The first element will be the pathname of the executable.
	 *
	 *	Bugs: tends to return the MS-DOS form of the application path.
	 *	Todo: implement on platforms other than Windows.
	 */
	string[] parseCommandLine() {
		return parseCommandLine(GetCommandLineW());
	}

	/**	Parses a command _line given as a null-terminated UTF-16 string into an argument list.
	 *
	 *	Params:
	 *		line = the command _line.
	 *	Returns:
	 *		an array of the arguments.
	 *
	 *	Todo: implement on platforms other than Windows (if necessary).
	 */
	string[] parseCommandLine(const(wchar)* line) {
		return parseCommandLine(wcharzToUTF8(line));
	}

} else {
	/**	Converts a single argument from application form to command line form.
	 *
	 *	Todo: finish implementing on platforms other than Windows.  (How to literalise !, \ and '?)
	 */
	string formatArgument(string arg) {
		if (countChars(arg, "`" `!"$&*();'|\<>/? `) == 0) {
			return arg;
		} else {
			return `'` ~ arg ~ `'`;
		}
	}
}
