/* ******************************************************************* *\
*                               console.d                               *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*      Console input/output with support for Windows character sets     *
>***********************************************************************<
* Copyright (C) 2006-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
The smjg.libs.util.console module provides a means of communicating with the standard input and
output streams.

$(P It is designed to provide the following benefits over stream classes in Phobos:)
$(UL
$(LI support Windows character sets)
$(LI fix a bug in Win9x* whereby a line of output is lost on stdin EOF)
$(LI flag EOF of stdin immediately from the class user's point of view.)
)

$(P &#42; At this time of writing, programs compiled with DMD2 don't run under Win9x, but the code
          of relevance is still there just in case.)

$(P The application programmer would use this module by means of the cIn, cOut and cErr objects.)

$(P This is not intended to be a sophisticated text I/O library.)

$(P Any application that uses this module should use it consistently throughout if possible.  This
ensures that the EOF fix works correctly and prevents buffering conflicts between different I/O
libraries/APIs.)

Since: 0.03
Todo:  implement error checking (not sure how much is needed)
*/
module smjg.libs.util.console;

private import
	std.conv,
	std.exception,
	std.format,
	std.string,
	std.traits,
	std.utf,
	std.c.stdio,
	std.c.windows.windows,
	std.windows.syserror;


version (Windows) {
	private import std.windows.charset;

	extern (C) int isatty(int);

	extern (Windows) {
		//int GetConsoleCP();
		//int GetConsoleOutputCP();
		HANDLE GetStdHandle(uint);
		BOOL ReadConsoleW(HANDLE, LPVOID, DWORD, LPDWORD, LPVOID);
	}
}


/**	The ConsoleInput class is used to receive text from the standard input.
 *
 *	$(P This class buffers input on a line-by-line basis.)
 */
class ConsoleInput {
	/**	Reads a single Unicode character from the input stream.
	 *
	 *	Returns: the character retrieved, or dchar.init on EOF.
	 */
	dchar readChar() {
		if (eof) return dchar.init;

		return decode(buffer, bufPos);
	}

	/**	Discards the currently buffered line of input.
	 *
	 *	$(P The next call to readChar, readLine or eof will therefore wait for
	 *	a line of input from the user and use it.)
	 *
	 *	If no text is waiting in the buffer, this function has no effect.
	 */
	void discardLine() {
		bufPos = buffer.length = 0;
	}

	/**	The number of UTF-8 bytes currently waiting in the internal buffer.
	 *
	 *	$(P For as long as this property returns non-zero, input operations will be non-blocking.)
	 */
	@property size_t bytesInBuffer() {
		return buffer.length - bufPos;
	}

	/**	The number of UTF-8 bytes remaining on the current line of input.
	 *
	 *	$(P This differs from $(D bytesInBuffer) in that the count doesn't include any terminating
	 *	newline.  But it isn't equivalent to $(D bytesInBuffer - 1): if the buffer is empty, or the
	 *	input stream has been buffered to EOF without a final newline, then both properties will
	 *	return the same value.)
	 */
	@property size_t bytesInLine() {
		size_t result = buffer.length - bufPos;
		if (result != 0 && buffer[$-1] == '\n') {
			return result - 1;
		} else {
			return result;
		}
	}

	///	The number of characters currently waiting in the internal buffer.
	@property size_t charsInBuffer() {
		return toUCSindex(buffer[bufPos..$], buffer.length - bufPos);
	}

	///	The number of characters remaining on the current line of input.
	@property size_t charsInLine() {
		size_t result = charsInBuffer;
		if (result != 0 && buffer[$-1] == '\n') {
			return result - 1;
		} else {
			return result;
		}
	}

	/**	Reads a line of text from the input stream.
	 *
	 *	Params:
	 *		withNewLine = whether to include in the returned string the terminating newline if it
	 *		is present.
	 *	Returns:
	 *		the line retrieved, or \c null on EOF.
	 */
	char[] readLine(bool withNewLine = false) {
		if (eof) return null;

		scope(exit) bufPos = buffer.length = 0;

		if (!withNewLine && buffer[$-1] == '\n') {
			return buffer[bufPos..$-1].dup;
		} else {
			return buffer[bufPos..$].dup;
		}
	}

	/**	Determines whether EOF of the input stream has been reached.
	 *
	 *	$(P This function also reads a line of input into the buffer if it is empty and there is
	 *	any input to read.  Therefore, it will wait for the user to input a line if one isn't
	 *	already available.)
	 *
	 *	Returns:
	 *		true if the end of the input has been reached and the buffer is empty, otherwise false.
	 */
	@property bool eof() {
		if (bufPos != buffer.length) return false;
		if (_eof) return true;

		// nothing left in buffer - read line
		buffer.length = 0;
		bufPos = 0;

		version (Windows) {
			if (useReadConsole) {
				wchar[] wBuffer;

				scope (success) {
					buffer = toUTF8(wBuffer);
				}

				while (true) {
					wchar c;
					uint wcharsRead;

					// TODO: improve exception
					if (!ReadConsoleW(handle, &c, 1, &wcharsRead, null)) {
						throw new ConsoleException("Read from console failed: "
						  ~ sysErrorString(GetLastError()));
					}
					if (wcharsRead != 1 || c == '\u001A') {
						_eof = 1;
						return wBuffer.length == 0;
					}

					switch (c) {
						case '\r':
							wBuffer ~= '\n';
							discardLF = true;
							return false;

						case '\n':
							if (!discardLF) {
								wBuffer ~= '\n';
								return false;
							}
							discardLF = false;
							break;

						default:
							discardLF = false;
							wBuffer ~= c;
					}
				}
			} else {
				scope (success) {
					auto bufz = toStringz(buffer);
					version (D_Version2) {
						buffer = fromMBSz(cast(immutable) bufz, GetConsoleCP());
					} else {
						buffer = fromMBSz(bufz, GetConsoleCP());
					}
				}

				while (true) {
					char c;
					uint bytesRead;

					// TODO: improve exception
					if (!ReadFile(handle, &c, 1, &bytesRead, null)) {
						throw new ConsoleException("Read from console failed: "
						  ~ sysErrorString(GetLastError()));
					}
					if (bytesRead != 1) {
						_eof = 1;
						return buffer.length == 0;
					}

					switch (c) {
						case '\r':
							buffer ~= '\n';
							discardLF = true;
							return false;

						case '\n':
							if (!discardLF) {
								buffer ~= '\n';
								return false;
							}
							discardLF = false;
							break;

						default:
							discardLF = false;
							buffer ~= c;
					}
				}
			}

		} else {
			while (true) {
				int c = getc(stream);

				switch (c) {
					case '\r': case '\n':
						buffer ~= '\n';
						return false;

					case -1:
						_eof = 1;
						return buffer.length == 0;

					default:
						buffer ~= cast(char) c;
				}
			}
		}
		assert (false);
	}

	private {
		string buffer;
		size_t bufPos;
		FILE* stream;
		/*	The EOF state of the input stream.  Values:
		 *	0 - not EOF
		 *	1 - EOF, line may be lost
		 *	2 - EOF, lost line recovered
		 */
		int _eof = 0;
		version (Windows) {
			HANDLE handle;
			bool discardLF;

			/*	Whether to use the Windows API function ReadConsoleW.
			 *
			 *	This is important because:
			 *	\li ReadConsoleW is supported only on Windows NT/2000 and above
			 *	\li ReadConsoleW requires that the input is actually a console and not
			 *		redirected
			 *	\li the alternative, ReadFile, is bitten by a Windows bug if it tries to
			 *		read a Unicode character from the console.
			 */
			bool useReadConsole;
		}

		this() {
			stream = stdin;

			version (Windows) {
				handle = GetStdHandle(-10);

				useReadConsole = !(GetVersion() & 0x80000000)
				  && isatty(stdin._file);
			}
		}
	}

	invariant() {
		assert (bufPos <= buffer.length);
		//if (_eof) assert (buffer.length == 0);
	}
}


/**	The ConsoleOutput class used to write out text to the standard output and error streams.
 *
 *	$(P This class buffers any incomplete UTF-8 characters that are passed into it.  This makes it
 *	suited to applications that transmit UTF-8 text one byte at a time, or in chunks whose
 *	boundaries may not coincide with character boundaries.  It is the application's responsibility
 *	to ensure that the sequence of strings written constitutes well-formed UTF-8 text.)
 */
class ConsoleOutput {
	/**	Writes out the Unicode character c.
	 *
	 *	Precondition: There must be no partial UTF-8 characters remaining in the buffer.
	 */
	void write(dchar c)
	in {
		assert (bufPos == buffer.length);
	}
	body {
		buffer.length = bufPos = 0;
		encode(buffer, c);
		flushBuffer();
	}
	
	version (Windows) {
		///	Writes out the string c.
		void write(const(char)[] c) {
			buffer ~= c;
			flushBuffer();
		}

		///	Writes out the string c followed by a newline.
		void writeLine(const(char)[] c) {
			buffer ~= c ~ '\n';
			flushBuffer();
		}

	} else {
		///	Writes out the string c.
		void write(const(char)[] str) {
			if (bufPos != buffer.length) {
				buffer ~= str;
				flushBuffer();
			} else {
				fwrite(str.ptr, 1, str.length, stream);
			}
		}

		///	Writes out the string c followed by a newline.
		void writeLine(const(char)[] str) {
			if (bufPos != buffer.length) {
				buffer ~= str ~ '\n';
				flushBuffer();
			} else {
				fwrite(str.ptr, 1, str.length, stream);
				putc('\n', stream);
			}
		}
	}

	///	Performs formatted output on the stream.
	void writef(S...)(S args) if (isSomeString!(S[0])) {
		formattedWrite(this, args);
		flushBuffer();
	}
	
	/// ditto
	void writefln(S...)(S args) if (isSomeString!(S[0])) {
		formattedWrite(this, args);
		buffer ~= '\n';
		flushBuffer();
	}
	
	/**	Deprecated: designed for the old std.string.doFormat way of doing things;
	 *	use writef/writefln instead.
	 */
	deprecated void writefx(TypeInfo[] arguments, void* argptr, bool newline = false)
	in {
		assert (bufPos == buffer.length);
	} body {
		doFormat(delegate void(dchar c) { encode(buffer, c); }, arguments, argptr);
		if (newline) buffer ~= '\n';
		flushBuffer();
	}

	// helper functions for writef*
	void put(char c) { buffer ~= c; }
	void put(wchar c) { encode(buffer, c); }
	void put(dchar c) { encode(buffer, c); }
	void put(const(char)[] s) { buffer ~= s; }
	void put(const(wchar)[] s) { foreach (c; s) encode(buffer, c); }
	void put(const(dchar)[] s) { foreach (c; s) encode(buffer, c); }

	///	Makes sure all pending output has actually been output.
	void flush() {
		flushBuffer();
		fflush(stream);
	}

	private {
		char[] buffer;
		size_t bufPos;
		FILE* stream;

		version (Windows) {
			HANDLE handle;

			//!	Whether to fix the bug in Win9x whereby a line of output is lost
			//!	when stdin reaches EOF.
			/*!
			 *	\defaultValue  true for Win9x; false for NT/2000 and above.  If
			 *	               you discover that this is wrong for any Windows
			 *	               version from 95/NT4 upwards, please fix the
			 *	               processing of GetVersion accordingly.
			 */
			bool fixEofBug;
		}

		this(FILE* s)
		in {
			assert (s == stdout || s == stderr);
		}
		body {
			stream = s;

			version (Windows) {
				handle = GetStdHandle(s == stdout ? -11 : -12);

				fixEofBug = (GetVersion() & 0x80000000)
				  && isatty(stdin._file) && isatty(s._file);
			}
		}

		version (Windows) {
			void flushBuffer() {
				// first check if an extra line of output is needed
				if (fixEofBug && cIn._eof == 1) {
					cIn._eof = 2;
					putc('\n', stream);
				}

				// find last whole character
				char expected;
				size_t newBufPos, bytesWritten;

				foreach_reverse (index, char c; buffer) {
					if (!(c & 0x80)) {
						newBufPos = index + 1;
						break;
					} else if (!(c & 0x40)) {
						if (expected == char.init) {
							expected = 0xC0;
						} else {
							expected = cast(char) ((expected >> 1) | 0x80);
						}
					} else {
						assert ((c & 0xC0) == 0xC0);
						if ((c & expected) == expected) {
							newBufPos = buffer.length;
						} else {
							newBufPos = index;
						}
						break;
					}
				}
				
				string toWrite = to!string
				  (toMBSz(buffer[bufPos..newBufPos], GetConsoleOutputCP()));

				WriteFile(handle, toWrite.ptr, toWrite.length, &bytesWritten, null);
				bufPos = newBufPos;
				
				if (bufPos == buffer.length) bufPos = buffer.length = 0;
			}

		} else {
			void flushBuffer() {
				fwrite(&buffer[bufPos], 1, buffer.length - bufPos, stream);
				bufPos = buffer.length = 0;
			}
		}
	}

	invariant() {
		assert (bufPos <= buffer.length);
	}
}


///	The standard input stream.
ConsoleInput cIn;

///	The standard output stream.
ConsoleOutput cOut;

///	The standard error stream.
ConsoleOutput cErr;


static this() {
	cIn = new ConsoleInput();
	cOut = new ConsoleOutput(stdout);
	cErr = new ConsoleOutput(stderr);
}


///	Exception thrown when a console I/O operation fails.
class ConsoleException : Exception {
	this(string msg) { super(msg); }
}
