/* ******************************************************************* *\
*                               hashmap.d                               *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*    Hash table implementation of an associative array, which doesn't   *
*          assume that the key type has an ordering comparator          *
>***********************************************************************<
* Copyright (C) 2005-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
This module provides a hash map container template.
*/
module smjg.libs.util.hashmap;

///	Convenience alias for smjg.libs.util.hashmap.THashMap!(K, V).Map.
template HashMap(K, V) {
	alias THashMap!(K, V).Map HashMap;
}


/**	Future_Directions:
 *		The map class will likely be defined directly as HashMap, with iterator or range classes
 *		as nested classes.  THashMap will then be deprecated.
 */
template THashMap(K, V) {
	private struct Node {
		K key;
		V value;
	}

	/**	The Map class implements an associative container.
	 *
	 *	$(P Unlike the current implementation of D's built-in associative arrays, this doesn't
	 *	rely on an ordering comparator, so it will work even if the key type has no ordering or
	 *	unequal objects can rank equally in order.)
	 *
	 *	Future_Directions:
	 *		Map will enforce immutability of the key type.
	 *		An abstract base class similar to Set will probably be set up for generality.
	 */
	class Map {
		public {
			///	Returns: the value with key k, or V.init if the key is not present.
			V opIndex(K k) const {
				Node[]* slot = typeid(K).getHash(&k) in hashTable;

				if (slot == null) return V.init;

				foreach (Node n; *slot) {
					if (n.key == k) {
						return n.value;
					}
				}

				return V.init;
			}

			///	Sets the value with key k to be v.
			V opIndexAssign(V v, K k) {
				hash_t hash = typeid(K).getHash(&k);
				Node[]* slot = hash in hashTable;

				if (slot == null) {
					Node[] newNode = new Node[1];
					with (newNode[0]) {
						key = k;
						value = v;
					}
					hashTable[hash] = newNode;
					return v;
				}

				foreach (inout Node n; *slot) {
					if (n.key == k) {
						n.value = v;
						return v;
					}
				}

				slot.length = slot.length + 1;
				with ((*slot)[$ - 1]) {
					key = k;
					value = v;
				}
				return v;
			}

			///	Returns: whether the map contains key k.
			bool opIn_r(K k) {
				Node[]* slot = typeid(K).getHash(&k) in hashTable;
				if (slot == null) return false;
				//Node[] slot = hashTable[k.toHash];

				foreach (Node n; *slot) {
					if (n.key == k) {
						return true;
					}
				}

				return false;
			}

			/**	Removes the key k.
			 *	
			 *	Returns:
			 *		the value that was associated with this key, or V.init if k it was not present.
			 */
			V remove(K k) {
				Node[]* slot = &hashTable[typeid(K).getHash(&k)];

				if (slot == null) return V.init;

				foreach (inout Node n; *slot) {
					if (n.key == k) {
						V result = n.value;
						// move last item to fill space
						if (n !is (*slot)[slot.length - 1]) {
							n = (*slot)[slot.length - 1];
						}

						slot.length = slot.length - 1;
						return result;
					}
				}
			}

			///	Returns: the number of keys stored in the map.
			size_t length() {
				size_t result = 0;

				foreach (slot; hashTable) result += slot.length;
				return result;
			}

			/**	Rehashes the array in place.
			 *	
			 *	Returns: the array.
			 */
			Map rehash() {
				Map newMap = new Map;

				foreach (K key, V value; this) {
					newMap[key] = value;
				}

				newMap.hashTable.rehash;
				hashTable = newMap.hashTable;
				return this;
			}


			/**	Iterates over the (keys and) values in the map.
			 *
			 *	$(P Modifying the map while foreach is in progress, other than by a ref value
			 *	variable in the foreach loop, currently leads to undefined behaviour.)
			 */
			int opApply(int delegate(ref V) dg) {
				int result = 0;

				foreach (uint hash, Node[] slot; hashTable) {
					foreach (ref Node node; slot) {
						result = dg(node.value);
						if (result) break;
					}
				}
				return result;
			}

			/// ditto
			int opApply(int delegate(ref const(V)) dg) const {
				int result = 0;

				foreach (uint hash, Node[] slot; hashTable) {
					foreach (ref const(Node) node; slot) {
						result = dg(node.value);
						if (result) break;
					}
				}
				return result;
			}

			/// ditto
			int opApply(int delegate(ref const(K), ref V) dg) {
				int result = 0;

				foreach (uint hash, Node[] slot; hashTable) {
					foreach (inout Node node; slot) {
						result = dg(node.key, node.value);
						if (result) break;
					}
				}
				return result;
			}

			/// ditto
			int opApply(int delegate(ref const(K), ref const(V)) dg) const {
				int result = 0;

				foreach (uint hash, Node[] slot; hashTable) {
					foreach (inout Node node; slot) {
						result = dg(node.key, node.value);
						if (result) break;
					}
				}
				return result;
			}
		}

		private Node[][uint] hashTable;
	}


	/**	The MapIterator class is used to iterate over the elements of a HashMap.
	 *	
	 *	$(P Modifying the map, other than by the iterator's value property, and then continuing to
	 *	use the iterator, currently leads to undefined behaviour.)
	 */
	class MapIterator {
		public {
			///	Constructs a new iterator over m.
			this(Map m) {
				data = m.hashTable.values;
				while (!atEnd && itemNo == data[hashNo].length) {
					hashNo++;
					itemNo = 0;
				}
			}

			/**	The key associated with the iterator's current position.
			 *
			 *	Precondition: The iterator must not be at the end.
			 */
			@property K key()
			in {
				assert (!atEnd);
			}
			body {
				return data[hashNo][itemNo].key;
			}

			/**	The value associated with the iterator's current position.
			 *
			 *	Precondition: The iterator must not be at the end.
			 */
			@property V value()
			in {
				assert (!atEnd);
			}
			body {
				return data[hashNo][itemNo].value;
			}

			/// ditto
			@property V value(V v)
			in {
				assert (!atEnd);
			}
			body {
				data[hashNo][itemNo].value = v;
				return v;
			}

			///	Advances the iterator.
			void opPostInc()
			in {
				assert (!atEnd);
			}
			body {
				itemNo++;
				while (!atEnd && itemNo == data[hashNo].length) {
					hashNo++;
					itemNo = 0;
				}
			}

			/**	Returns:
			 *		whether the iterator has reached the terminal position following the final
			 *		key in the map.
			 */
			bool atEnd() {
				return hashNo == data.length;
			}
		}

		private {
			Node[][] data;
			uint hashNo = 0, itemNo = 0;
		}
	}
}
