/* ******************************************************************* *\
*                               setbase.d                               *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.10                             *
*                                                                       *
*            Abstract base class for implementations of a set           *
>***********************************************************************<
* Copyright (C) 2006-2012 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\* ******************************************************************* */
/**
This module provides an implementation of an abstract base class for set classes.

Since: 0.03
*/
module smjg.libs.util.setbase;

private import smjg.libs.util.types;
private import std.string: format;
debug import std.stdio;


/**	The Set class is the abstract base class from which set classes can be derived.
 *
 *	Default implementations of several methods are given.  These may be overridden to provide an
 *	implementation optimised to the internal data structure used by the specific set
 *	implementation.
 */
abstract class Set(Element) if (isImmutable!(Element)) {
	public {
		abstract {
			/**	Adds elem to the set.
			 *
			 *	Bugs:
			 *		As of DMD 2.058, it is possible to add a mutable object of a subclass of
			 *		Element's set ($(BZ 3731)).
			 */
			void add(Element elem);

			///	Removes elem from the set.
			void remove(Element elem);

			/**	Removes the element currently active in the innermost active foreach loop on
			 *	this set.
			 *
			 *	Precondition:
			 *		A foreach loop must be active on this set.
			 *	
			 *	Warning:
			 *		The time at which the removal takes effect may depend on
			 *	    the implementation: it might take effect immediately, or not until the
			 *	    iteration of the foreach loop exits.  It is also possible that, if within
			 *	    one foreach loop another is initiated on the same set after a call to
			 *		removeCurrent, the wrong element may be removed.  There is also potential
			 *		undefined behaviour if the same set is iterated through a mutable
			 *		reference and, nested within this loop, through a const or immutable
			 *		reference.
			 *	
			 *	Future_Directions:
			 *		This method may be superseded by something more robust.
			 */
			void removeCurrent();
			
			///	Returns: the set's copy of elem, or Element.init if the element is not present.
			Element storedKey(Element elem) const;

			///	Returns: the number of elements in the set.
			@property size_t length() const;

			///	Returns: an exact copy of this set.
			@property Set dup() const;

			///	Returns: an exact immutable copy of this set.
			@property immutable(Set) idup() const;

			///	Returns: an array of the set's elements.
			@property Element[] data() const;

			///	Returns: whether elem is present in the set.
			bool opIn_r(Element elem) const;

			/**	Iterates over the elements of the set.
			 *
			 *	$(P Modifying the set while foreach is in progress, except by the removeCurrent
			 *	method, may lead to undefined behaviour.)
			 */
			int opApply(int delegate(ref Element) dg);
			/// ditto
			int opApply(int delegate(ref Element) dg) const;
		}

		///	Returns: whether this set is a subset of s.
		bool subsetOf(const(Set) s) const {
			foreach (ref Element e; this) {
				if (!s[e]) return false;
			}
			return true;
		}

		/**	Removes from this set all elements for which dg returns true.
		 *
		 *	Returns: a set of the elements removed.
		 */
		Set remove(bool delegate(Element) dg) {
			Set removed = this & dg;
			this -= removed;
			return removed;
		}

		/**	Removes from this set all elements that are in s.
		 *
		 *	Returns: a set of the elements removed.
		 */
		Set remove(const(Set) s) {
			Set removed = this & s;
			this -= removed;
			return removed;
		}

		/**	Returns:
		 *		a string of the form { element, element, ... } where the elements are formatted
		 *		according to the %s format specifier.
		 *
		 *	To_Do: change to const once $(BZ 1824) is fixed
		 */
		string toString() /+const+/ {
			string result = "{";
			bool anyElem = false;

			foreach (ref Element e; this) {
				result ~= format(anyElem ? ", %s" : "%s", e);
				anyElem = true;
			}
			result ~= '}';
			return result;
		}

		/**	Returns:
		 *		the hash value of this set, calculated by XORing together the hash values of all
		 *		the elements.
		 *	To_Do: change to const once $(BZ 1824) is fixed
		 */
		hash_t toHash() /+const+/ {
			hash_t result;

			foreach (ref Element e; this) {
				result ^= typeid(Element).getHash(&e);
			}
			return result;
		}

		override {
			// workaround for Bugzilla 1824
			/+string toString() { return (cast(const) this).toString(); }
			hash_t toHash() { return (cast(const) this).toHash(); }+/

			///	Returns: whether this set is equal to other.
			bool opEquals(Object other) {
				Set s = cast(Set) other;
				if (s is null) return false;

				return length == s.length && subsetOf(s);
			}
		}

		/// ditto
		bool opEquals(const(Set) other) const {
			return length == other.length && subsetOf(other);
		}

		///
		alias opIn_r opIndex;

		/**	Adds or removes an element to or from this set.
		 *
		 *	Params:
		 *		elem = the element to add or remove.
		 *		b    = true to add the element, or false to remove it.
		 *	Returns: b.
		 */
		bool opIndexAssign(bool b, Element elem) {
			if (b) {
				add(elem);
			} else {
				remove(elem);
			}
			return b;
		}

		///	Returns: the intersection of this set and s.
		Set opAnd(const(Set) s) const {
			return dup.opAndAssign(s);
		}

		///	Returns: the union of this set and s.
		Set opOr(const(Set) s) const {
			return dup.opOrAssign(s);
		}

		///	Returns: a set containing the elements of this set that are not in s.
		Set opSub(const(Set) s) const {
			return dup.opSubAssign(s);
		}

		///	Returns: the symmetric difference between this set and s.
		Set opXor(const(Set) s) const {
			return dup.opXorAssign(s);
		}

		///	Returns: a set containing the elements of this set that for which dg returns true.
		Set opAnd(bool delegate(Element) dg) const {
			return dup.opAndAssign(dg);
		}

		///	Returns: a set containing the elements of this set that for which dg returns false.
		Set opSub(bool delegate(Element) dg) const {
			return dup.opSubAssign(dg);
		}

		///
		Set opAndAssign(const(Set) s) {
			foreach (ref Element e; this) {
				if (!(e in s)) removeCurrent();
			}
			return this;
		}

		///
		Set opOrAssign(const(Set) s) {
			foreach (ref Element e; s) add(e);
			return this;
		}

		///
		Set opSubAssign(const(Set) s) {
			foreach (ref Element e; s) remove(e);
			return this;
		}

		///
		Set opXorAssign(const(Set) s) {
			Element[] newElements;

			foreach (ref Element e; s) this[e] = !this[e];
			return this;
		}

		///
		Set opAndAssign(bool delegate(Element) dg) {
			//debug puts("entered opAndAssign");
			foreach (ref Element e; this) {
				if (!dg(e)) {
					//debug writefln("removing %s", e);
					removeCurrent();
				} else {
					//debug writefln("keeping %s", e);
				}
			}
			return this;
		}

		///
		Set opSubAssign(bool delegate(Element) dg) {
			foreach (ref Element e; this) {
				if (dg(e)) removeCurrent();
			}
			return this;
		}
	}
}

unittest {
	import std.stdio;

	Set!(immutable(int)) set1;
	Set!(immutable(Set!(immutable(int)))) set2;
	
	puts("SetBase passed!");
}
